# from useless.pipes import worker
import inspect 
from useless.pipes.pipes import Worker as _Worker
from functools import partial, wraps, update_wrapper

class Incompatible(Exception): pass

anything = '*'

class Worker(_Worker):
    def __ror__(self, left):
        try:
            yields = left.yields()
            if not all( (self.accepts(symbol) for symbol in yields.split()) ):
                raise Incompatible('Incompatible left: %s to right: %s' % (yields, self.accepts()))
        except AttributeError:
            pass

        return super(Worker, self).__ror__(left)

def _get_positional_arguments(func):
    spec = inspect.getargspec(func)
    args_with_defaults = spec.defaults and len(spec.defaults) or 0
    return spec.args[:-args_with_defaults or None] 

def _worker(func):
    positional_args = _get_positional_arguments(func)
    # when we have only the mandatory first argument which is the iterator on the left side
    # allowing multiple binds would lead to an infinite bind-loop, we would never get a working worker
    allow_multiple_binds = positional_args > 1

    @wraps(func)
    def bind(*a, **kw):
        # trigger three step process if kw has only keyword arguments
        # in python we can def f(a, b=1) => f(a=1, b=2)
        # t.i. a positional arg can be treated as if it were a kw argument
        # but this shouldn't trigger three-step, only e.g. f(b=2)
        if allow_multiple_binds and kw and not a and not set(kw).intersection(positional_args):
            return wraps(bind) (partial(bind, *a, **kw))

        @wraps(bind)
        def apply(iter):
            return apply.original_function(iter, *a, **kw)

        return wraps(apply) (Worker(apply))

    bind.original_function = func

    def decorate_with(f):
        bind.original_function = f(bind.original_function)
    bind.decorate_with = decorate_with

    return bind


def worker(func=None, accepts=anything, yields=anything):
    def _accepts(symbol=None):
        return (symbol in accepts or anything in [accepts, symbol]) if symbol else accepts
    def _yields(symbol=None):
        return (symbol in yields or yields == anything) if symbol else yields

    if func:
        func.accepts = _accepts
        func.yields = _yields
        return _worker(func)


    @wraps(worker)
    def wrapped(f):
        return worker(f, accepts, yields)

    return wrapped   

filter = worker

def discover_filters(module):
    """Helper for the usual

            __all__ = discover_filters(globals())
        
    """
    return [symbol for symbol, code in module.iteritems() if getattr(code, 'original_function', False)]




import os
import hashlib
import urlparse

from __init__ import FilterError

import ass.ets 
import ass.ets.bundles
from ass.ets.workers import filter, Incompatible, discover_filters

@filter(accepts='filenames', yields='filenames')
def local_path(files, bundle):
	"""Expects relative paths and yields absolute paths using bundle.map_from."""
	for file in files:
		yield os.path.join(bundle.map_from, file)


@filter(accepts='filenames', yields='filenames')
def remote_path(files, bundle):
	"""Expects relative paths and yields urls by using bundle.map_to."""
	for file in files:
		base = bundle.map_to
		base += '/' if base[-1:] != '/' else ''

		yield urlparse.urljoin(base, file) 

@filter(accepts='filenames', yields='filenames')
def translate_path(files, bundle):
	"""Given an absolute path, yields urls. 
	"""
	return files | relative_path(bundle) | remote_path(bundle)

@filter(accepts='filenames', yields='filenames')
def relative_path(files, bundle, root=None):
	"""Given an absolute path yields a relative one. By default uses
	bundle.map_from as the root.
	"""
	root = root or bundle.map_from

	for file in files:
		yield os.path.relpath(file, root)

@filter
def echo(items, bundle):
	for item in items:
		yield item

@filter(accepts='filenames bundles', yields='filenames')
def as_is(files, bundle):
	"""Just echoes filenames, but for nested bundles yields what they yield.
	"""
	for file in files:
		if isinstance(file, ass.ets.Bundle):
			nested_bundle = file
			iterator = nested_bundle.apply()
			if not iterator.yields('filenames'):
				raise Incompatible("%r must yield 'filenames', actually yields %r" % (nested_bundle, iterator.yields()))
			# handling nested bundles still feels hacky
			# assume the nested bundle yields relative paths with a root = subbundle.map_from
			# first localize, so we have an absolute path
			# then translate that path again to a relative path with the root = bundle.map_from
			for f in iterator | local_path(nested_bundle) | relative_path(bundle):
				yield f
		else:
			yield file

@filter(yields='filenames')
def ask_manifest(files, bundle, key=None):
	"""Assumes a Manifest on bundle.manifest which answers get(key). 
	By default the bundle's name is used as the key.
	Yields one or more local paths.
	"""
	# we need to consume the given iterator
	for file in files: pass

	key = key or bundle.name
	for file in bundle.manifest.get(key):
		yield file

use_manifest = ask_manifest

@filter(accepts='filenames', yields='filenames')
def store_manifest(files, bundle, key=None):
	"""Assumes a Manifest on bundle.manifest which answers set(key, value).
	By default the bundle's name is used as the key. The value is a 
	list of local paths.
	"""
	filenames = []
	for file in files:
		filenames.append(file)
		yield file

	key = key or bundle.name
	bundle.manifest.set(bundle.name, filenames)

@filter(accepts='filenames bundles', yields='contents')
def read(items, bundle):
	"""Reads the files and yields their contents. For nested bundles, just yields
	whatever they yield.  
	"""
	for item in items:
		if isinstance(item, ass.ets.Bundle):
			nested_bundle = item
			iterator = nested_bundle.apply()
			if not iterator.yields('contents'):
				raise Incompatible("%r must yield 'contents', actually yields %r" % (nested_bundle, iterator.yields()))
			for content in iterator:
				yield content
		else:
			filename = os.path.join(bundle.map_from, item)
			with open(filename) as file:
				yield file.read()


@filter(accepts='contents', yields='contents')
def merge(contents, bundle):
	"""Merge the contents of files. 

		['a', 'b'] | merge(bundle=None) == ['ab']


	"""
	yield ''.join(contents)

@filter(accepts='contents', yields='filenames')
def store(contents, bundle, name=None):
	"""Writes the given contents to disc. By default uses the name stored
	on bundle.output. 
	A '%(version)s' tag in the name gets replaced by hashing the content. 
	"""
	name = name or bundle.output
	versioned = '%(version)s' in name
	
	for content in contents:
		filename = name
		if versioned:
			hash = hashlib.md5(content).hexdigest()[:8]
			filename = filename % dict(version=hash) 

		full_path = os.path.join(bundle.map_from, filename)
		with open(full_path, 'wb') as file:
			file.write(content)

		yield filename 

def store_as(name):
	"""The same as calling store(name=name)"""
	return store(name=name)

import glob as glob_
@filter(accepts='filenames bundles', yields='filenames bundles')
def glob(files, bundle):
	for file in files:
		if isinstance(file, ass.ets.Bundle):
			yield file
		else:
			full_path = os.path.join(bundle.map_from, file)
			for expanded in glob_.iglob(full_path) | relative_path(bundle):
				yield expanded



# 

from ass.ets.dicts import ordered

def _get_pipe_for(ext, bundle):
	return ass.ets.bundles.Pipe( bundle.filters[ext][bundle.mode] )

@filter(accepts='filenames')
def automatic(files, bundle):
	"""Assumes a dict on bundle.filters, like so::

		env = Environment(
			filters={
				'js': {
					'development': [read, merge]
				},
				'css': {
					'development': [read, merge, minify]
				}
			}
		)
		bundle = Bundle(
			'jquery.js', 'styles.css',
			env=env,
			# partly overwrite the above defaults
			filters={ 'js': { 'development': [read, merge, uglifyjs] } },
			development=automatic
		)

	The automatic filter then chooses the correct filter chain by looking
	at the extension of the file.

	Alpha quality: Does not handle nested bundles.
	"""
	by_ext = ordered()
	for file in files:
		_, ext = os.path.splitext(file)
		ext = ext[1:]
		try:
			by_ext[ext].append(file)
		except KeyError:
			by_ext[ext] = [file]

	for ext, files in by_ext.iteritems():
		pipe = _get_pipe_for(ext, bundle)
		for thing in pipe.apply(files, bundle):
			yield thing


__all__ = discover_filters(globals()) + ['store_as',]



class option_property(property): pass

def option(name, desc='', getter=None, setter=None, default=None):
	def get(self):
		try:
			return self._options[name]
		except KeyError:
			try:
				return getattr(self.parent, name)
			except AttributeError:
				if default:
					return default

				raise AttributeError("%r is undefined" % name)

	def set(self, value):
		self._options[name] = value

	def del_(self):
		self._options.pop(name)

	# print setter, set, setter or set
	return option_property(getter and getter(name, default) or get, 
					setter and setter(name) or set, 
					del_, desc)

def pipelize_getter(name, default=None):
	def getter(self):
		try:
			rv = self._options[name]
		except KeyError:
			try:
				rv = getattr(self.parent, name)
			except AttributeError:
				if default:
					rv = default

				raise AttributeError("%s is undefined." % name)

		return Pipe(rv)

	return getter



def dict_getter(name, default=None):

	def _update(parent, this):
		for k, v in this.iteritems():
			if parent.has_key(k) and isinstance(v, dict):
				_update(parent[k], v)
			else:
				parent[k] = v
		
	def getter(self):
		this = self._options.get(name, default or {})
		parent = getattr( getattr(self, 'parent', None), name, {}).copy()

		if not parent and not this:	
			raise Undefined("%s is undefined." % name)

		_update(parent, this)

		return parent

	return getter



class Option(object):
	def __init__(self, name=None, **kw):
		self.name = name
		self.kw = kw

	def get_property(self):
		return option(self.name, **self.kw)

# we inject this as a @property at self._options
def _get_options_(self):
	try:
		return self.__options
	except AttributeError:
		self.__options = dict()
		return self.__options

class InjectOptions(type):
	"""Puts a dict on self._options and a option-property for each defined Option"""
	def __init__(cls, name, bases, dict_):
		super(InjectOptions, cls).__init__(name, bases, dict_)
		setattr(cls, '_options', property(_get_options_))

		for key, option in dict_.iteritems():
			if isinstance(option, Option):
				if option.name is None:
					option.name = key
				setattr(cls, option.name, option.get_property())

class Options(object):
	__metaclass__ = InjectOptions

	def __new__(cls, *a, **kw):
		instance = super(Options, cls).__new__(cls)

		for key, value in kw.iteritems():
			varname = getattr(cls, key, None)
			if isinstance(varname, option_property):
				setattr(instance, key, value)

		return instance




import os, warnings, operator

try:
	import yaml as serializer
except ImportError:
	import pickle as serializer

from options import Option, Options, dict_getter
import filters as f
import dicts

def manifest_setter(name):
	def setter(self, manifest):
		if isinstance(manifest, str):
			manifest = os.path.join(self.map_from, manifest)
			manifest = Manifest(manifest)

		self._options[name] = manifest

	return setter

class Manifest(object):
	"""Naive implementation. Uses yaml or pickle to serialize.

	"""
	def __init__(self, filename, serializer=serializer):
		self.filename = filename
		self.serializer = serializer
		self._manifest = None

	@property
	def manifest(self):
		if self._manifest is None:
			try:
				with open(self.filename) as f:
					self._manifest = serializer.load(f)
			except:
				self._manifest = {}
		return self._manifest

	def get(self, key):
		return self.manifest[key]

	def set(self, key, value):
		self.manifest[key] = value
		self._save_manifest()

	def _save_manifest(self):
		with open(self.filename, 'wb') as f:
			serializer.dump(self.manifest, f)

			
class CommonOptions(Options):
	map_from = Option()
	map_to = Option(default='/')
	mode = Option()
	manifest = Option(setter=manifest_setter)
	filters = Option(getter=dict_getter)

	# because of the options-inheritance we define these here, even though
	# we might only need them way down for the bundles
	production = Option()
	development = Option()#getter=pipelize_getter)
	build_ = Option()

	def __init__(self, env=None, **kw):
		if env is not None:
			self.env = env

	@property
	def parent(self):
		return self.env


class bundleslist(dicts.listdicthybrid, dicts.dotted):
	__slots__ = ('data', '_env', '_key_func_')

	def __init__(self, data=[], env=None):
		self._env = env
		super(bundleslist, self).__init__(data, key_func=lambda i: getattr(i, 'name', None))

	def _prepare_bundle(self, bundle, key):
		if key is not None:
			try:
				getattr(bundle, 'name')
			except:
				bundle.name = key
		try:
			getattr(bundle, 'env')
		except:
			bundle.env = self._env

	def append(self, item):
		self._prepare_bundle(item, None)
		super(bundleslist, self).append(item)

	def __setitem__(self, key, value):
		self._prepare_bundle(value, key)
		super(bundleslist, self).__setitem__(key, value)

class assetslist(bundleslist):
	#assets can be either just a path or a bundle 
	def _prepare_bundle(self, bundle, key):
		if not isinstance(bundle, Bundle):
			return
		super(assetslist, self)._prepare_bundle(bundle, key)

	def _index_of(self, key):
		for i, v in enumerate(self.data):
			try:
				if self._key_func_(v) == key:
					return i
			except AttributeError:
				pass
		else:
			raise KeyError

class Pipe(list):
    def __init__(self, seq):
        if not hasattr(seq, '__iter__'):
            seq = [seq]

        list.__init__(self, seq)

    def accepts(self, symbol=None):
        return self[0].accepts(symbol)

    def yields(self, symbol=None):
        return self[-1].yields(symbol)

    def prepend(self, worker):
        self.insert(0, worker)

    def apply(self, iter, *a, **kw):
        p = iter
        for worker in self:
            p |= worker(*a, **kw)

        return p

class Environment(CommonOptions): 
	"""An Environment is just a configuration object. Consider:

		env = Environment(map_from='./static')
		js = bundle(env=env)

		js.map_from == './static'


	"""

class Assets(CommonOptions):
	"""An Assets-instance is a container for bundles. The
	name is probably confusing.

		assets = Assets(
			bundle(name='bundleA', assets=['fileA'...] ...),
			production=use_manifest, #etc
		)

		# now you could:
		for bundle in assets: ....
			bundle.build()

		# or access the named bundle, just like
		assets.bundles.bundleA.build()

		# or assign a bundle
		assets.bundles.javascripts = bundle(...)

	Notice, that the name of the bundle, t.i. its name-attribute, gets
	used as the key for the bundles-mapping.

	"""
	def __init__(self, *bundles, **kw):
		super(Assets, self).__init__(**kw)
		if not bundles and kw.has_key('bundles'):
			bundles = kw['bundles']
		self.bundles = bundles

	@property
	def bundles(self):
		return self._bundles

	@bundles.setter
	def bundles(self, new_value):
		if not isinstance(new_value, bundleslist):
			new_value = bundleslist(new_value, env=self)
		self._bundles = new_value

	def __iter__(self):
		return self.bundles.itervalues()

	def __repr__(self):
		return "Assets(%s)" % ', '.join(map(lambda i: "%r" % i, self.bundles.values()))


class Bundle(CommonOptions):
	"""A Bundle bundles files or nested bundles. Named bundles
	can be accessed via the assets member.

		less_style = bundle(name=less)
		styles = bundle(
			less_style, 'main.css', #...
		)

		styles.assets.less == less_style
		styles.assets['less'] == less_style

	Bundles provide the two useful methods urls() and build() and
	the underlying method apply().

	Usually you set the mode and call urls() and or build()

		styles.make_it = [read, merge] #...
		styles.mode = 'make_it'
		styles.build()

	But better start with the mode names 'development', 'production' 
	and 'build_'.


	"""
	name = Option()
	output = Option()

	def __init__(self, *assets, **kw):
		super(Bundle, self).__init__(**kw)
		if not assets and kw.has_key('assets'):
			assets = kw['assets']
		self.assets = assets

	@property
	def assets(self):
		return self._assets

	@assets.setter
	def assets(self, new_value):
		if not isinstance(new_value, assetslist):
			new_value = assetslist(new_value, env=self)
		self._assets = new_value



	def urls(self, urlize=f.remote_path):
		"""Applies the mode self.mode, converts paths to urls.
		
		Actually returns a generator you have to consume.

		"""
		return self.apply(append=urlize)

	def apply(self, mode=None, pipe=None, append=None):
		"""Either provide a list of workers/filters on pipe or
		a name of a mode, otherwise uses self.mode.

		Actually not necessarily applies the filters but returns
		a generator you have to consume.  
		"""
		pipe = Pipe( pipe or getattr(self, mode or self.mode) )
		if append:
			pipe.append(append)

		return pipe.apply(self.assets, self)

	def build(self, mode=None, localize=f.local_path):
		"""Applies the set mode, converts relative paths to absolute 
		local paths and returns a list.

		"""
		if mode:
			warnings.warn('Because of the way nested bundles are built, passing a mode to build() might not work as you might expect.')

		return self.apply(mode=mode, append=localize) | list

	def __repr__(self):
		try:
			return "Bundle(name=%r)" % self.name
		except:
			return "Bundle(%s)" % ', '.join(map(lambda i: "%r" % i, self.assets.values()))


bundle = Bundle


import subprocess
import sys
on_windows = sys.platform == 'win32'

from __init__ import FilterError
from ass.ets.workers import filter, Incompatible, discover_filters

@filter(accepts='contents', yields='contents')
def popens(files, bundle, args=None, shell=True if on_windows else False, name=None):
	"""Standard subprocess.Popen that expects a pipe on stdin and stdout.
	"""
	assert args is not None
	name = name or args[0] # assume we have a good name on the first argument which is the binary

	for content in files:
		proc = subprocess.Popen(
			args,
			stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
			shell=shell)
		stdout, stderr = proc.communicate(content)

		if proc.returncode != 0:
			raise FilterError(('%s: subprocess had error: stderr=%s, '+
                               'stdout=%s, returncode=%s') % (
                                    name, stderr, stdout, proc.returncode))

		yield stdout
	
uglifyjs = popens(args=['uglifyjs'])
lessify  = popens(args=['lessc', '-'])
cleancss = popens(args=['cleancss'])

@filter(accepts='contents', yields='contents')
def decaffeinate(files, bundle, bin='coffee', bare=False):
	args = [bin, '-sp' + 'b' if bare else '']
	return files | popens(bundle, args=args)


@filter(accepts='contents', yields='contents')
def cssminify(files, bundle):
	import cssmin

	for file in files:
		yield cssmin.cssmin(file)

minify = cssminify

__all__ = discover_filters(globals())



"""
    ass.ets
    
    Asset management for webapps.

    :copyright: (c) 2012 by Herrn Kaste <herr.kaste@gmail.com>.
    :license: BSD, see LICENSE for more details.
"""

__version__ = "0.1.1"

from bundles import Environment, Assets, bundle, Bundle, Manifest
from filters import FilterError
import filters as f
from options import Option, Options, dict_getter
from ass.ets.workers import worker, filter, Incompatible






class listdicthybrid(object):
	"""Holds a list of objects. Provides some limited dict-like access."""
	
	def __init__(self, data=[], key_func=None):
		self.data = []
		self._key_func_ = key_func
		if isinstance(data, dict):
			self.update(data)
		else:
			self.extend(data)

	def extend(self, iter):
		for i in iter:
			self.append(i)

	def append(self, item):
		self.data.append( item )

	def update(self, d):
		for k, v in d.iteritems():
			self[k] = v

	def _index_of(self, key):
		for i, v in enumerate(self.data):
			if self._key_func_(v) == key:
				return i
		else:
			raise KeyError
		
	def __getitem__(self, key):
		try:
			i = self._index_of(key)
		except KeyError:
			raise
		return self.data[i]

	def __setitem__(self, key, bundle):
		try:
			i = self._index_of(key)
		except KeyError:
			return self.append( bundle )
		self.data[i] = bundle

	def __delitem__(self, key):
		try:
			i = self._index_of(key)
		except KeyError:
			raise IndexError
		self.data.pop(i)

	def values(self):
		return self.data

	def keys(self):
		return filter(None, map(self._key_func_, self.data))

	def __iter__(self):
		return iter(self.data)

	def __repr__(self):
		return repr(self.values())
		

class ordered(object):
	"""Naive implementaion of an ordered dict."""
	def __init__(self, data={}):
		self.data = []
		self.update(data)

	def update(self, d):
		for k, v in d.items():
			self[k] = v

	def __getitem__(self, key):
		for k, v in self.data:
			if k == key:
				return v
		else:				
			raise KeyError

	def __setitem__(self, key, value):
		for i, (k, v) in enumerate(self.data):
			if k == key:
				self.data[i] = (key, value)
				break 
		else:
			self.data.append((key, value))

	def __delitem__(self, key):
		for i, (k, v) in enumerate(self.data):
			if k == key:
				self.data.pop(i)
				break
		else:
			raise IndexError

	def iterkeys(self):
		return (k for k, _ in self.data)

	def itervalues(self):
		return (v for _, v in self.data)

	def iteritems(self):
		return iter(self.data)

	def __iter__(self):
		return self.iterkeys()

class dotted(object):
	"""Naive mixin for a dict to enable attribute-style access."""
	__slots__ = ('data',)
	def __getattr__(self, key):
		try:
			return self.__getitem__(key)
		except KeyError:
			raise AttributeError

	def __setattr__(self, key, value):
		if key in self.__class__.__slots__:
			object.__setattr__(self, key, value)
		else:
			self.__setitem__(key, value)

	def __delattr__(self, key):
		if key in self.__class__.__slots__:
			object.__delattr__(self, key)
		else:
			self.__delitem__(key)
	
